
#include <stdlib.h>
#include "ha.h"


/**********************************************************************
	Delta Preprocessor
***********************************************************************/

S32B testsamp8bit(unsigned char *data, U32B len)
{
	S32B sum8, dlt8;
	char dat8a, dat8b;
	U32B i;

	sum8 = 0;

	dat8a = 0;

	for(i=len>>1;i>0;i--)
	{
		dat8b = (char)(data[0]);
		dlt8 = (dat8b - dat8a);
		sum8 += abs(dat8b) - abs(dlt8);
		data += 1;

		dat8a = (char)(data[0]);
		dlt8 = (dat8a - dat8b);
		sum8 += abs(dat8a) - abs(dlt8);
		data += 1;
	}

	return sum8;
}

S32B testsamp16be(unsigned char *data, U32B len)
{
	S32B sum16b, dlt16;
	char dat16a, dat16b;
	U32B i;

	sum16b = 0;

	dat16a = 0;

	for(i=len>>2;i>0;i--)
	{
		dat16b = (char)(data[0]);
		dlt16 = (dat16b - dat16a);
		sum16b += abs(dat16b) - abs(dlt16);
		data += 2;

		dat16a = (char)(data[0]);
		dlt16 = (dat16a - dat16b);
		sum16b += abs(dat16a) - abs(dlt16);
		data += 2;
	}

	return sum16b*2;
}

S32B testsamp16le(unsigned char *data, U32B len)
{
	S32B sum16l, dlt16;
	char dat16a, dat16b;
	U32B i;

	sum16l = 0;

	dat16a = 0;

	for(i=len>>2;i>0;i--)
	{
		dat16b = (char)(data[1]);
		dlt16 = (dat16b - dat16a);
		sum16l += abs(dat16b) - abs(dlt16);
		data += 2;

		dat16a = (char)(data[1]);
		dlt16 = (dat16a - dat16b);
		sum16l += abs(dat16a) - abs(dlt16);
		data += 2;
	}

	return sum16l*2;
}


void dodelta8bit(unsigned char *src, unsigned char *dst, U32B len)
{
	char dat8a, dat8b, dlt8;
	U32B i;

	dat8a = 0;
	dat8b = 0;

	if (len&1)
	{
		dat8a = (char)(src[0]);
		dlt8 = (char)(dat8a - dat8b);
		dst[0] = (unsigned char)(dlt8);
		src += 1;
		dst += 1;
	}

	for(i=len>>1;i>0;i--)
	{
		dat8b = (char)(src[0]);
		dlt8 = (char)(dat8b - dat8a);
		dst[0] = (unsigned char)(dlt8);
		src += 1;
		dst += 1;

		dat8a = (char)(src[0]);
		dlt8 = (char)(dat8a - dat8b);
		dst[0] = (unsigned char)(dlt8);
		src += 1;
		dst += 1;
	}
}

void undodelta8bit(unsigned char *src, unsigned char *dst, U32B len)
{
	char dat8;
	U32B i;

	dat8 = 0;

	if (len&1)
	{
		dat8 += (char)(src[0]);
		dst[0] = (unsigned char)(dat8);
		src += 1;
		dst += 1;
	}

	for(i=len>>1;i>0;i--)
	{
		dat8 += (char)(src[0]);
		dst[0] = (unsigned char)(dat8);
		src += 1;
		dst += 1;

		dat8 += (char)(src[0]);
		dst[0] = (unsigned char)(dat8);
		src += 1;
		dst += 1;
	}
}

/* delta without de-interleaving */
#if 0
void dodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	S16B dat16a, dat16b, dlt16;
	U32B i;

	dat16a = 0;
	dat16b = 0;

	if (len&2)
	{
		dat16a = (S16B)((src[0]<<8)|src[1]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16>>8);
		dst[1] = (unsigned char)(dlt16&255);
		src += 2;
		dst += 2;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b = (S16B)((src[0]<<8)|src[1]);
		dlt16 = (S16B)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16>>8);
		dst[1] = (unsigned char)(dlt16&255);
		src += 2;
		dst += 2;

		dat16a = (S16B)((src[0]<<8)|src[1]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16>>8);
		dst[1] = (unsigned char)(dlt16&255);
		src += 2;
		dst += 2;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void undodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	S16B dat16;
	U32B i;

	dat16 = 0;

	if (len&2)
	{
		dat16 += (S16B)((src[0]<<8)|src[1]);
		dst[0] = (unsigned char)(dat16>>8);
		dst[1] = (unsigned char)(dat16&255);
		src += 2;
		dst += 2;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16 += (S16B)((src[0]<<8)|src[1]);
		dst[0] = (unsigned char)(dat16>>8);
		dst[1] = (unsigned char)(dat16&255);
		src += 2;
		dst += 2;

		dat16 += (S16B)((src[0]<<8)|src[1]);
		dst[0] = (unsigned char)(dat16>>8);
		dst[1] = (unsigned char)(dat16&255);
		src += 2;
		dst += 2;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void dodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	S16B dat16a, dat16b, dlt16;
	U32B i;

	dat16a = 0;
	dat16b = 0;

	if (len&2)
	{
		dat16a = (S16B)((src[1]<<8)|src[0]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16&255);
		dst[1] = (unsigned char)(dlt16>>8);
		src += 2;
		dst += 2;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b = (S16B)((src[1]<<8)|src[0]);
		dlt16 = (S16B)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16&255);
		dst[1] = (unsigned char)(dlt16>>8);
		src += 2;
		dst += 2;

		dat16a = (S16B)((src[1]<<8)|src[0]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16&255);
		dst[1] = (unsigned char)(dlt16>>8);
		src += 2;
		dst += 2;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void undodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	S16B dat16;
	U32B i;

	dat16 = 0;

	if (len&2)
	{
		dat16 += (S16B)((src[1]<<8)|src[0]);
		dst[0] = (unsigned char)(dat16&255);
		dst[1] = (unsigned char)(dat16>>8);
		src += 2;
		dst += 2;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16 += (S16B)((src[1]<<8)|src[0]);
		dst[0] = (unsigned char)(dat16&255);
		dst[1] = (unsigned char)(dat16>>8);
		src += 2;
		dst += 2;

		dat16 += (S16B)((src[1]<<8)|src[0]);
		dst[0] = (unsigned char)(dat16&255);
		dst[1] = (unsigned char)(dat16>>8);
		src += 2;
		dst += 2;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}
#endif

/* delta (combined high/low-byte) with de-interleaving */
#if 0
void dodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	S16B dat16a, dat16b, dlt16;
	U32B i;

	ptr = &dst[(len+1)>>1];

	dat16a = 0;
	dat16b = 0;

	if (len&2)
	{
		dat16a = (S16B)((src[0]<<8)|src[1]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16>>8);
		ptr[0] = (unsigned char)(dlt16&255);
		src += 1;
		dst += 1;
		src += 1;
		ptr += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b = (S16B)((src[0]<<8)|src[1]);
		dlt16 = (S16B)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16>>8);
		ptr[0] = (unsigned char)(dlt16&255);
		src += 1;
		dst += 1;
		src += 1;
		ptr += 1;

		dat16a = (S16B)((src[0]<<8)|src[1]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16>>8);
		ptr[0] = (unsigned char)(dlt16&255);
		src += 1;
		dst += 1;
		src += 1;
		ptr += 1;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void undodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	S16B dat16;
	U32B i;

	ptr = &src[(len+1)>>1];

	dat16 = 0;

	if (len&2)
	{
		dat16 += (S16B)((src[0]<<8)|ptr[0]);
		dst[0] = (unsigned char)(dat16>>8);
		dst[1] = (unsigned char)(dat16&255);
		src += 1;
		dst += 1;
		ptr += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16 += (S16B)((src[0]<<8)|ptr[0]);
		dst[0] = (unsigned char)(dat16>>8);
		dst[1] = (unsigned char)(dat16&255);
		src += 1;
		dst += 1;
		ptr += 1;
		dst += 1;

		dat16 += (S16B)((src[0]<<8)|ptr[0]);
		dst[0] = (unsigned char)(dat16>>8);
		dst[1] = (unsigned char)(dat16&255);
		src += 1;
		dst += 1;
		ptr += 1;
		dst += 1;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void dodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	S16B dat16a, dat16b, dlt16;
	U32B i;

	ptr = &dst[len>>1];

	dat16a = 0;
	dat16b = 0;

	if (len&2)
	{
		dat16a = (S16B)((src[1]<<8)|src[0]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16>>8);
		ptr[0] = (unsigned char)(dlt16&255);
		src += 1;
		ptr += 1;
		src += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b = (S16B)((src[1]<<8)|src[0]);
		dlt16 = (S16B)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16>>8);
		ptr[0] = (unsigned char)(dlt16&255);
		src += 1;
		ptr += 1;
		src += 1;
		dst += 1;

		dat16a = (S16B)((src[1]<<8)|src[0]);
		dlt16 = (S16B)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16>>8);
		ptr[0] = (unsigned char)(dlt16&255);
		src += 1;
		ptr += 1;
		src += 1;
		dst += 1;
	}

	if (len&1)
	{
		ptr[0] = src[0];
	}
}

void undodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	S16B dat16;
	U32B i;

	ptr = &src[len>>1];

	dat16 = 0;

	if (len&2)
	{
		dat16 += (S16B)((src[0]<<8)|ptr[0]);
		dst[0] = (unsigned char)(dat16&255);
		dst[1] = (unsigned char)(dat16>>8);
		ptr += 1;
		dst += 1;
		src += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16 += (S16B)((src[0]<<8)|ptr[0]);
		dst[0] = (unsigned char)(dat16&255);
		dst[1] = (unsigned char)(dat16>>8);
		ptr += 1;
		dst += 1;
		src += 1;
		dst += 1;

		dat16 += (S16B)((src[0]<<8)|ptr[0]);
		dst[0] = (unsigned char)(dat16&255);
		dst[1] = (unsigned char)(dat16>>8);
		ptr += 1;
		dst += 1;
		src += 1;
		dst += 1;
	}

	if (len&1)
	{
		dst[0] = ptr[0];
	}
}
#endif

/* delta (high-byte only) with de-interleaving */
#if 1
void dodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16a, dat16b, dlt16;
	U32B i;

	ptr = &dst[(len+1)>>1];

	dat16a = 0;
	dat16b = 0;

	if (len&2)
	{
		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
		ptr[0] = src[0];
		src += 1;
		ptr += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b = (char)(src[0]);
		dlt16 = (char)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
		ptr[0] = src[0];
		src += 1;
		ptr += 1;

		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
		ptr[0] = src[0];
		src += 1;
		ptr += 1;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void undodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16;
	U32B i;

	ptr = &src[(len+1)>>1];

	dat16 = 0;

	if (len&2)
	{
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
		dst[0] = ptr[0];
		ptr += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
		dst[0] = ptr[0];
		ptr += 1;
		dst += 1;

		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
		dst[0] = ptr[0];
		ptr += 1;
		dst += 1;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void dodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16a, dat16b, dlt16;
	U32B i;

	ptr = &dst[len>>1];

	dat16a = 0;
	dat16b = 0;

	if (len&2)
	{
		ptr[0] = src[0];
		src += 1;
		ptr += 1;
		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		ptr[0] = src[0];
		src += 1;
		ptr += 1;
		dat16b = (char)(src[0]);
		dlt16 = (char)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;

		ptr[0] = src[0];
		src += 1;
		ptr += 1;
		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
	}

	if (len&1)
	{
		ptr[0] = src[0];
	}
}

void undodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16;
	U32B i;

	ptr = &src[len>>1];

	dat16 = 0;

	if (len&2)
	{
		dst[0] = ptr[0];
		ptr += 1;
		dst += 1;
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dst[0] = ptr[0];
		ptr += 1;
		dst += 1;
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;

		dst[0] = ptr[0];
		ptr += 1;
		dst += 1;
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
	}

	if (len&1)
	{
		dst[0] = ptr[0];
	}
}
#endif

/* delta (separate high/low-byte) with de-interleaving */
#if 0
void dodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16a, dat16b, dlt16;
	char dat16a2, dat16b2;
	U32B i;

	ptr = &dst[(len+1)>>1];

	dat16a = 0;
	dat16b = 0;

	dat16a2 = 0;
	dat16b2 = 0;

	if (len&2)
	{
		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
		dat16a2 = (char)(src[0]);
		dlt16 = (char)(dat16a2 - dat16b2);
		ptr[0] = (unsigned char)(dlt16);
		src += 1;
		ptr += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b = (char)(src[0]);
		dlt16 = (char)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
		dat16b2 = (char)(src[0]);
		dlt16 = (char)(dat16b2 - dat16a2);
		ptr[0] = (unsigned char)(dlt16);
		src += 1;
		ptr += 1;

		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
		dat16a2 = (char)(src[0]);
		dlt16 = (char)(dat16a2 - dat16b2);
		ptr[0] = (unsigned char)(dlt16);
		src += 1;
		ptr += 1;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void undodelta16be(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16;
	char dat162;
	U32B i;

	ptr = &src[(len+1)>>1];

	dat16 = 0;
	dat162 = 0;

	if (len&2)
	{
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
		dat162 += (char)(ptr[0]);
		dst[0] = (unsigned char)(dat162);
		ptr += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
		dat162 += (char)(ptr[0]);
		dst[0] = (unsigned char)(dat162);
		ptr += 1;
		dst += 1;

		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
		dat162 += (char)(ptr[0]);
		dst[0] = (unsigned char)(dat162);
		ptr += 1;
		dst += 1;
	}

	if (len&1)
	{
		dst[0] = src[0];
	}
}

void dodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16a, dat16b, dlt16;
	char dat16a2, dat16b2;
	U32B i;

	ptr = &dst[len>>1];

	dat16a = 0;
	dat16b = 0;

	dat16a2 = 0;
	dat16b2 = 0;

	if (len&2)
	{
		dat16a2 = (char)(src[0]);
		dlt16 = (char)(dat16a2 - dat16b2);
		ptr[0] = (unsigned char)(dlt16);
		src += 1;
		ptr += 1;
		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat16b2 = (char)(src[0]);
		dlt16 = (char)(dat16b2 - dat16a2);
		ptr[0] = (unsigned char)(dlt16);
		src += 1;
		ptr += 1;
		dat16b = (char)(src[0]);
		dlt16 = (char)(dat16b - dat16a);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;

		dat16a2 = (char)(src[0]);
		dlt16 = (char)(dat16a2 - dat16b2);
		ptr[0] = (unsigned char)(dlt16);
		src += 1;
		ptr += 1;
		dat16a = (char)(src[0]);
		dlt16 = (char)(dat16a - dat16b);
		dst[0] = (unsigned char)(dlt16);
		src += 1;
		dst += 1;
	}

	if (len&1)
	{
		ptr[0] = src[0];
	}
}

void undodelta16le(unsigned char *src, unsigned char *dst, U32B len)
{
	unsigned char *ptr;
	char dat16;
	char dat162;
	U32B i;

	ptr = &src[len>>1];

	dat16 = 0;
	dat162 = 0;

	if (len&2)
	{
		dat162 += (char)(ptr[0]);
		dst[0] = (unsigned char)(dat162);
		ptr += 1;
		dst += 1;
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
	}

	for(i=len>>2;i>0;i--)
	{
		dat162 += (char)(ptr[0]);
		dst[0] = (unsigned char)(dat162);
		ptr += 1;
		dst += 1;
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;

		dat162 += (char)(ptr[0]);
		dst[0] = (unsigned char)(dat162);
		ptr += 1;
		dst += 1;
		dat16 += (char)(src[0]);
		dst[0] = (unsigned char)(dat16);
		src += 1;
		dst += 1;
	}

	if (len&1)
	{
		dst[0] = ptr[0];
	}
}
#endif

